# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import lale.pretty_print

class TestToGraphviz(unittest.TestCase):
    def test_with_operator_choice(self):
        from lale.operators import make_choice
        from lale.lib.lale import NoOp
        from lale.lib.sklearn import KNeighborsClassifier
        from lale.lib.sklearn import LogisticRegression
        from lale.lib.sklearn import Nystroem
        from lale.lib.sklearn import PCA
        kernel_tfm_or_not =  NoOp | Nystroem
        tfm = PCA
        clf = make_choice(LogisticRegression, KNeighborsClassifier)
        clf.visualize(ipython_display=False)
        optimizable = kernel_tfm_or_not >> tfm >> clf
        optimizable.visualize(ipython_display=False)

    def test_invalid_input(self):
        from sklearn.linear_model import LogisticRegression as SklearnLR
        scikit_lr = SklearnLR()
        from lale.helpers import to_graphviz
        with self.assertRaises(TypeError):
            to_graphviz(scikit_lr)

class TestPrettyPrint(unittest.TestCase):
    def _roundtrip(self, expected, printed):
        self.maxDiff = None
        self.assertEqual(expected, printed)
        globals2 = {}
        exec(printed, globals2)
        pipeline2 = globals2['pipeline']

    def test_indiv_op_1(self):
        from lale.lib.sklearn import LogisticRegression
        pipeline = LogisticRegression(solver=LogisticRegression.enum.solver.saga, C=0.9)
        expected = """from lale.lib.sklearn import LogisticRegression
import lale
lale.wrap_imported_operators()

pipeline = LogisticRegression(solver='saga', C=0.9)"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_indiv_op_2(self):
        from lale.lib.sklearn import LogisticRegression
        pipeline = LogisticRegression()
        expected = """from lale.lib.sklearn import LogisticRegression
import lale
lale.wrap_imported_operators()

pipeline = LogisticRegression()"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_reducible(self):
        from lale.lib.sklearn import MinMaxScaler
        from lale.lib.lale import NoOp
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import Nystroem
        from lale.lib.lale import ConcatFeatures
        from lale.lib.sklearn import KNeighborsClassifier
        from lale.lib.sklearn import LogisticRegression
        pca = PCA(copy=False)
        logistic_regression = LogisticRegression(solver='saga', C=0.9)
        pipeline = (MinMaxScaler | NoOp) >> (pca & Nystroem) >> ConcatFeatures >> (KNeighborsClassifier | logistic_regression)
        expected = \
"""from lale.lib.sklearn import MinMaxScaler
from lale.lib.lale import NoOp
from lale.lib.sklearn import PCA
from lale.lib.sklearn import Nystroem
from lale.lib.lale import ConcatFeatures
from lale.lib.sklearn import KNeighborsClassifier
from lale.lib.sklearn import LogisticRegression
import lale
lale.wrap_imported_operators()

pca = PCA(copy=False)
logistic_regression = LogisticRegression(solver='saga', C=0.9)
pipeline = (MinMaxScaler | NoOp) >> (pca & Nystroem) >> ConcatFeatures >> (KNeighborsClassifier | logistic_regression)"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_no_combinators(self):
        from lale.lib.sklearn import MinMaxScaler
        from lale.lib.lale import NoOp
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import Nystroem
        from lale.lib.lale import ConcatFeatures
        from lale.lib.sklearn import KNeighborsClassifier
        from lale.lib.sklearn import LogisticRegression
        pca = PCA(copy=False)
        logistic_regression = LogisticRegression(solver='saga', C=0.9)
        pipeline = (MinMaxScaler | NoOp) >> (pca & Nystroem & NoOp) >> ConcatFeatures >> (KNeighborsClassifier | logistic_regression)
        expected = \
"""from lale.lib.sklearn import MinMaxScaler
from lale.lib.lale import NoOp
from lale.operators import make_choice
from lale.lib.sklearn import PCA
from lale.lib.sklearn import Nystroem
from lale.operators import make_union
from lale.lib.sklearn import KNeighborsClassifier
from lale.lib.sklearn import LogisticRegression
from lale.operators import make_pipeline

choice_0 = make_choice(MinMaxScaler, NoOp)
pca = PCA(copy=False)
union = make_union(pca, Nystroem, NoOp)
logistic_regression = LogisticRegression(solver='saga', C=0.9)
choice_1 = make_choice(KNeighborsClassifier, logistic_regression)
pipeline = make_pipeline(choice_0, union, choice_1)"""
        printed = lale.pretty_print.to_string(pipeline, combinators=False)
        self._roundtrip(expected, printed)

    def test_import_as_1(self):
        from lale.lib.sklearn import LogisticRegression as LR
        pipeline = LR(solver='saga', C=0.9)
        expected = """from lale.lib.sklearn import LogisticRegression as LR
import lale
lale.wrap_imported_operators()

pipeline = LR(solver='saga', C=0.9)"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_import_as_2(self):
        from lale.lib.sklearn import MinMaxScaler as Scaler
        from lale.lib.lale import NoOp
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import Nystroem
        from lale.lib.lale import ConcatFeatures as Concat
        from lale.lib.sklearn import KNeighborsClassifier as KNN
        from lale.lib.sklearn import LogisticRegression as LR
        pca = PCA(copy=False)
        lr = LR(solver='saga', C=0.9)
        pipeline = (Scaler | NoOp) >> (pca & Nystroem) >> Concat >> (KNN | lr)
        expected = \
"""from lale.lib.sklearn import MinMaxScaler as Scaler
from lale.lib.lale import NoOp
from lale.lib.sklearn import PCA
from lale.lib.sklearn import Nystroem
from lale.lib.lale import ConcatFeatures as Concat
from lale.lib.sklearn import KNeighborsClassifier as KNN
from lale.lib.sklearn import LogisticRegression as LR
import lale
lale.wrap_imported_operators()

pca = PCA(copy=False)
lr = LR(solver='saga', C=0.9)
pipeline = (Scaler | NoOp) >> (pca & Nystroem) >> Concat >> (KNN | lr)"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_operator_choice(self):
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import MinMaxScaler as Scl
        pipeline = PCA | Scl
        expected = \
"""from lale.lib.sklearn import PCA
from lale.lib.sklearn import MinMaxScaler as Scl
import lale
lale.wrap_imported_operators()

pipeline = PCA | Scl"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_higher_order(self):
        from lale.lib.lale import Both
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import Nystroem
        pipeline = Both(op1=PCA(n_components=2), op2=Nystroem)
        expected = """from lale.lib.lale import Both
from lale.lib.sklearn import PCA
from lale.lib.sklearn import Nystroem
import lale
lale.wrap_imported_operators()

pca = PCA(n_components=2)
pipeline = Both(op1=pca, op2=Nystroem)"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_higher_order_2(self):
        from lale.lib.sklearn import VotingClassifier as Vote
        from lale.lib.sklearn import KNeighborsClassifier as KNN
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import LogisticRegression as LR
        pipeline = Vote(estimators=[('knn',KNN), ('pipeline',PCA()>>LR)],
                        voting='soft')
        expected = """from lale.lib.sklearn import VotingClassifier as Vote
from lale.lib.sklearn import KNeighborsClassifier as KNN
from lale.lib.sklearn import PCA
from lale.lib.sklearn import LogisticRegression as LR
import lale
lale.wrap_imported_operators()

pipeline = Vote(estimators=[('knn', KNN), ('pipeline', PCA() >> LR)], voting='soft')"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_multimodal(self):
        from lale.lib.lale import Project
        from lale.lib.sklearn import Normalizer as Norm
        from lale.lib.sklearn import OneHotEncoder as OneHot
        from lale.lib.lale import ConcatFeatures as Cat
        from lale.lib.sklearn import LinearSVC
        project_0 = Project(columns={'type': 'number'})
        project_1 = Project(columns={'type': 'string'})
        linear_svc = LinearSVC(C=29617.4, dual=False, tol=0.005266)
        pipeline = ((project_0 >> Norm()) & (project_1 >> OneHot())) >> Cat >> linear_svc
        expected = \
"""from lale.lib.lale import Project
from lale.lib.sklearn import Normalizer as Norm
from lale.lib.sklearn import OneHotEncoder as OneHot
from lale.lib.lale import ConcatFeatures as Cat
from lale.lib.sklearn import LinearSVC
import lale
lale.wrap_imported_operators()

project_0 = Project(columns={'type': 'number'})
project_1 = Project(columns={'type': 'string'})
linear_svc = LinearSVC(C=29617.4, dual=False, tol=0.005266)
pipeline = ((project_0 >> Norm()) & (project_1 >> OneHot())) >> Cat >> linear_svc"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_irreducible_1(self):
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import Nystroem
        from lale.lib.sklearn import MinMaxScaler
        from lale.lib.sklearn import LogisticRegression
        from lale.lib.sklearn import KNeighborsClassifier
        from lale.operators import get_pipeline_of_applicable_type
        choice = PCA | Nystroem
        pipeline = get_pipeline_of_applicable_type(
            steps=[choice, MinMaxScaler, LogisticRegression, KNeighborsClassifier],
            edges=[(choice,LogisticRegression), (MinMaxScaler,LogisticRegression), (MinMaxScaler,KNeighborsClassifier)])
        expected = \
"""from lale.lib.sklearn import PCA
from lale.lib.sklearn import Nystroem
from lale.lib.sklearn import MinMaxScaler
from lale.lib.sklearn import LogisticRegression
from lale.lib.sklearn import KNeighborsClassifier
from lale.operators import get_pipeline_of_applicable_type
import lale
lale.wrap_imported_operators()

choice = PCA | Nystroem
pipeline = get_pipeline_of_applicable_type(steps=[choice, MinMaxScaler, LogisticRegression, KNeighborsClassifier], edges=[(choice,LogisticRegression), (MinMaxScaler,LogisticRegression), (MinMaxScaler,KNeighborsClassifier)])"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_irreducible_2(self):
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import MinMaxScaler as MMS
        from lale.lib.lale import ConcatFeatures as HStack
        from lale.lib.sklearn import KNeighborsClassifier as KNN
        from lale.lib.sklearn import LogisticRegression as LR
        from lale.operators import get_pipeline_of_applicable_type
        pipeline_0 = HStack >> LR
        pipeline = get_pipeline_of_applicable_type(
            steps=[PCA, MMS, KNN, pipeline_0],
            edges=[(PCA, KNN), (PCA, pipeline_0), (MMS, pipeline_0)])
        expected = \
"""from lale.lib.sklearn import PCA
from lale.lib.sklearn import MinMaxScaler as MMS
from lale.lib.sklearn import KNeighborsClassifier as KNN
from lale.lib.lale import ConcatFeatures as HStack
from lale.lib.sklearn import LogisticRegression as LR
from lale.operators import get_pipeline_of_applicable_type
import lale
lale.wrap_imported_operators()

pipeline_0 = HStack >> LR
pipeline = get_pipeline_of_applicable_type(steps=[PCA, MMS, KNN, pipeline_0], edges=[(PCA,KNN), (PCA,pipeline_0), (MMS,pipeline_0)])"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_nested(self):
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import LogisticRegression as LR
        from lale.lib.lale import NoOp
        lr_0 = LR(C=0.09)
        lr_1 = LR(C=0.19)
        pipeline = PCA >> (lr_0 | NoOp >> lr_1)
        expected = \
"""from lale.lib.sklearn import PCA
from lale.lib.sklearn import LogisticRegression as LR
from lale.lib.lale import NoOp
import lale
lale.wrap_imported_operators()

lr_0 = LR(C=0.09)
lr_1 = LR(C=0.19)
pipeline = PCA >> (lr_0 | NoOp >> lr_1)"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_autoai_libs_cat_encoder(self):
        from autoai_libs.transformers.exportable import CatEncoder
        import numpy as np
        from lale.lib.sklearn import LogisticRegression as LR
        cat_encoder = CatEncoder(categories='auto', dtype=np.float64, encoding='ordinal', handle_unknown='error')
        pipeline = cat_encoder >> LR()
        expected = \
"""from autoai_libs.transformers.exportable import CatEncoder
import numpy as np
from lale.lib.sklearn import LogisticRegression as LR
import lale
lale.wrap_imported_operators()

cat_encoder = CatEncoder(categories='auto', dtype=np.float64, encoding='ordinal', handle_unknown='error')
pipeline = cat_encoder >> LR()"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_autoai_libs_numpy_replace_missing_values(self):
        from autoai_libs.transformers.exportable import NumpyReplaceMissingValues
        from lale.lib.sklearn import LogisticRegression as LR
        numpy_replace_missing_values = NumpyReplaceMissingValues(
            filling_values=float('nan'), missing_values=['?'])
        pipeline = numpy_replace_missing_values >> LR()
        expected = \
"""from autoai_libs.transformers.exportable import NumpyReplaceMissingValues
from lale.lib.sklearn import LogisticRegression as LR
import lale
lale.wrap_imported_operators()

numpy_replace_missing_values = NumpyReplaceMissingValues(filling_values=float('nan'), missing_values=['?'])
pipeline = numpy_replace_missing_values >> LR()"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_autoai_libs_tam_1(self):
        from autoai_libs.cognito.transforms.transform_utils import TAM
        import autoai_libs.cognito.transforms.transform_extras
        import numpy as np
        from lale.lib.sklearn import LogisticRegression as LR
        tam = TAM(tans_class=autoai_libs.cognito.transforms.transform_extras.IsolationForestAnomaly, name='isoforestanomaly', col_names=['a', 'b', 'c'], col_dtypes=[np.dtype('float32'), np.dtype('float32'), np.dtype('float32')])
        pipeline = tam >> LR()
        expected = \
"""from autoai_libs.cognito.transforms.transform_utils import TAM
import autoai_libs.cognito.transforms.transform_extras
import numpy as np
from lale.lib.sklearn import LogisticRegression as LR
import lale
lale.wrap_imported_operators()

tam = TAM(tans_class=autoai_libs.cognito.transforms.transform_extras.IsolationForestAnomaly, name='isoforestanomaly', col_names=['a', 'b', 'c'], col_dtypes=[np.dtype('float32'), np.dtype('float32'), np.dtype('float32')])
pipeline = tam >> LR()"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))

    def test_autoai_libs_tam_2(self):
        from lale.lib.autoai_libs import TAM
        import numpy as np
        from lightgbm import LGBMClassifier
        from sklearn.decomposition import PCA
        from lale.operators import make_pipeline
        pca = PCA(copy=False)
        tam = TAM(tans_class=pca, name='pca', col_names=['a', 'b', 'c'], col_dtypes=[np.dtype('float32'), np.dtype('float32'), np.dtype('float32')])
        lgbm_classifier = LGBMClassifier(class_weight='balanced', learning_rate=0.18)
        pipeline = make_pipeline(tam, lgbm_classifier)
        import lale.helpers
        expected = \
"""from lale.lib.autoai_libs import TAM
import sklearn.decomposition.pca
import numpy as np
from lightgbm import LGBMClassifier
from lale.operators import make_pipeline

tam = TAM(tans_class=sklearn.decomposition.pca.PCA(copy=False, iterated_power='auto', n_components=None, random_state=None,   svd_solver='auto', tol=0.0, whiten=False), name='pca', col_names=['a', 'b', 'c'], col_dtypes=[np.dtype('float32'), np.dtype('float32'), np.dtype('float32')])
lgbm_classifier = LGBMClassifier(class_weight='balanced', learning_rate=0.18)
pipeline = make_pipeline(tam, lgbm_classifier)"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline, combinators=False))

    def test_autoai_libs_ta1(self):
        from autoai_libs.cognito.transforms.transform_utils import TA1
        import numpy as np
        import autoai_libs.utils.fc_methods
        from lale.lib.sklearn import LogisticRegression as LR
        ta1 = TA1(fun=np.rint, name='round', datatypes=['numeric'], feat_constraints=[autoai_libs.utils.fc_methods.is_not_categorical], col_names=['a', 'b', 'c'], col_dtypes=[np.dtype('float32'), np.dtype('float32'), np.dtype('float32')])
        pipeline = ta1 >> LR()
        expected = \
"""from autoai_libs.cognito.transforms.transform_utils import TA1
import numpy as np
import autoai_libs.utils.fc_methods
from lale.lib.sklearn import LogisticRegression as LR
import lale
lale.wrap_imported_operators()

ta1 = TA1(fun=np.rint, name='round', datatypes=['numeric'], feat_constraints=[autoai_libs.utils.fc_methods.is_not_categorical], col_names=['a', 'b', 'c'], col_dtypes=[np.dtype('float32'), np.dtype('float32'), np.dtype('float32')])
pipeline = ta1 >> LR()"""
        self._roundtrip(expected, lale.pretty_print.to_string(pipeline))


class TestToAndFromJSON(unittest.TestCase):
    def test_trainable_individual_op(self):
        self.maxDiff = None
        from lale.json_operator import to_json, from_json
        from lale.lib.sklearn import LogisticRegression as LR
        operator = LR(LR.solver.sag, C=0.1)
        json_expected = {
            'class': 'lale.lib.sklearn.logistic_regression.LogisticRegressionImpl',
            'state': 'trainable',
            'operator': 'LogisticRegression', 'label': 'LR',
            'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.logistic_regression.html',
            'hyperparams': {'C': 0.1, 'solver': 'sag'},
            'is_frozen_trainable': False}
        json = to_json(operator)
        self.assertEqual(json, json_expected)
        operator_2 = from_json(json)
        json_2 = to_json(operator_2)
        self.assertEqual(json_2, json_expected)

    def test_operator_choice(self):
        self.maxDiff = None
        from lale.json_operator import to_json, from_json
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import MinMaxScaler as Scl
        operator = PCA | Scl
        json_expected = {
          'class': 'lale.operators.OperatorChoice',
          'operator': 'OperatorChoice',
          'state': 'planned',
          'steps': {
            'pca': {
              'class': 'lale.lib.sklearn.pca.PCAImpl',
              'state': 'planned',
              'operator': 'PCA', 'label': 'PCA',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.pca.html'},
            'scl': {
              'class': 'lale.lib.sklearn.min_max_scaler.MinMaxScalerImpl',
              'state': 'planned',
              'operator': 'MinMaxScaler', 'label': 'Scl',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.min_max_scaler.html'}}}
        json = to_json(operator)
        self.assertEqual(json, json_expected)
        operator_2 = from_json(json)
        json_2 = to_json(operator_2)
        self.assertEqual(json_2, json_expected)

    def test_pipeline_1(self):
        self.maxDiff = None
        from lale.json_operator import to_json, from_json
        from lale.lib.lale import ConcatFeatures, NoOp
        from lale.lib.sklearn import LogisticRegression as LR
        from lale.lib.sklearn import PCA
        operator = (PCA & NoOp) >> ConcatFeatures >> LR
        json_expected = {
          'class': 'lale.operators.PlannedPipeline',
          'state': 'planned',
          'edges': [
              ['pca', 'concat_features'],
              ['no_op', 'concat_features'],
              ['concat_features', 'lr']],
          'steps': {
            'pca': {
              'class': 'lale.lib.sklearn.pca.PCAImpl',
              'state': 'planned',
              'operator': 'PCA', 'label': 'PCA',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.pca.html'},
            'no_op': {
              'class': 'lale.lib.lale.no_op.NoOpImpl',
              'state': 'trained',
              'operator': 'NoOp', 'label': 'NoOp',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.no_op.html',
              'hyperparams': None,
              'coefs': None,
              'is_frozen_trainable': True, 'is_frozen_trained': True},
            'concat_features': {
              'class': 'lale.lib.lale.concat_features.ConcatFeaturesImpl',
              'state': 'trained',
              'operator': 'ConcatFeatures', 'label': 'ConcatFeatures',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.concat_features.html',
              'hyperparams': None,
              'coefs': None,
              'is_frozen_trainable': True, 'is_frozen_trained': True},
            'lr': {
              'class': 'lale.lib.sklearn.logistic_regression.LogisticRegressionImpl',
              'state': 'planned',
              'operator': 'LogisticRegression', 'label': 'LR',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.logistic_regression.html'}}}
        json = to_json(operator)
        self.assertEqual(json, json_expected)
        operator_2 = from_json(json)
        json_2 = to_json(operator_2)
        self.assertEqual(json, json_2)

    def test_pipeline_2(self):
        from lale.lib.lale import NoOp
        from lale.lib.sklearn import Nystroem
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import LogisticRegression
        from lale.lib.sklearn import KNeighborsClassifier
        from lale.operators import make_choice, make_pipeline
        from lale.json_operator import to_json, from_json
        kernel_tfm_or_not =  make_choice(NoOp, Nystroem)
        tfm = PCA
        clf = make_choice(LogisticRegression, KNeighborsClassifier)
        operator = make_pipeline(kernel_tfm_or_not, tfm, clf)
        json = to_json(operator)
        operator_2 = from_json(json)
        json_2 = to_json(operator_2)
        self.assertEqual(json, json_2)

    def test_higher_order_1(self):
        from lale.lib.lale import Both
        from lale.lib.sklearn import PCA, Nystroem
        from lale.json_operator import from_json
        operator = Both(op1=PCA(n_components=2), op2=Nystroem)
        json_expected = {
          'class': 'lale.lib.lale.both.BothImpl',
          'state': 'trainable',
          'operator': 'Both', 'label': 'Both',
          'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.both.html',
          'hyperparams': {
            'op1': {'$ref': '../steps/pca'},
            'op2': {'$ref': '../steps/nystroem'}},
          'steps': {
            'pca': {
              'class': 'lale.lib.sklearn.pca.PCAImpl',
              'state': 'trainable',
              'operator': 'PCA', 'label': 'PCA',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.pca.html',
              'hyperparams': {'n_components': 2},
              'is_frozen_trainable': False},
            'nystroem': {
              'class': 'lale.lib.sklearn.nystroem.NystroemImpl',
              'state': 'planned',
              'operator': 'Nystroem', 'label': 'Nystroem',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.nystroem.html'}},
          'is_frozen_trainable': False}
        json = operator.to_json()
        self.assertEqual(json, json_expected)
        operator_2 = from_json(json)
        json_2 = operator_2.to_json()
        self.assertEqual(json, json_2)

    def test_higher_order_2(self):
        self.maxDiff = None
        from lale.lib.sklearn import VotingClassifier as Vote
        from lale.lib.sklearn import KNeighborsClassifier as KNN
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import LogisticRegression as LR
        from lale.json_operator import from_json
        operator = Vote(estimators=[('knn',KNN), ('pipeline',PCA()>>LR)],
                        voting='soft')
        json_expected = {
          'class': 'lale.lib.sklearn.voting_classifier.VotingClassifierImpl',
          'state': 'trainable',
          'operator': 'VotingClassifier',
          'is_frozen_trainable': True,
          'label': 'Vote',
          'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.voting_classifier.html',
          'hyperparams': {
            'estimators': [
              ('knn', {'$ref': '../steps/knn'}),
              ('pipeline', {'$ref': '../steps/pipeline'})],
            'voting': 'soft'},
          'steps': {
            'knn': {
              'class': 'lale.lib.sklearn.k_neighbors_classifier.KNeighborsClassifierImpl',
              'state': 'planned',
              'operator': 'KNeighborsClassifier', 'label': 'KNN',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.k_neighbors_classifier.html'},
            'pipeline': {
              'class': 'lale.operators.PlannedPipeline',
              'state': 'planned',
              'edges': [['pca', 'lr']],
              'steps': {
                'pca': {
                  'class': 'lale.lib.sklearn.pca.PCAImpl',
                  'state': 'trainable',
                  'operator': 'PCA', 'label': 'PCA',
                  'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.pca.html',
                  'hyperparams': {},
                  'is_frozen_trainable': False},
                'lr': {
                  'class': 'lale.lib.sklearn.logistic_regression.LogisticRegressionImpl',
                  'state': 'planned',
                  'operator': 'LogisticRegression', 'label': 'LR',
                  'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.logistic_regression.html'}}}}}
        json = operator.to_json()
        self.assertEqual(json, json_expected)
        operator_2 = from_json(json)
        json_2 = operator_2.to_json()
        self.assertEqual(json, json_2)

    def test_nested(self):
        self.maxDiff = None
        from lale.json_operator import to_json, from_json
        from lale.lib.lale import NoOp
        from lale.lib.sklearn import LogisticRegression as LR
        from lale.lib.sklearn import PCA
        operator = PCA >> (LR(C=0.09) | NoOp >> LR(C=0.19))
        json_expected = {
          'class': 'lale.operators.PlannedPipeline',
          'state': 'planned',
          'edges': [['pca', 'choice']],
          'steps': {
            'pca': {
              'class': 'lale.lib.sklearn.pca.PCAImpl',
              'state': 'planned',
              'operator': 'PCA', 'label': 'PCA',
              'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.pca.html'},
            'choice': {
              'class': 'lale.operators.OperatorChoice',
              'state': 'planned',
              'operator': 'OperatorChoice',
              'steps': {
                'lr_0': {
                  'class': 'lale.lib.sklearn.logistic_regression.LogisticRegressionImpl',
                  'state': 'trainable',
                  'operator': 'LogisticRegression', 'label': 'LR',
                  'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.logistic_regression.html',
                  'hyperparams': {'C': 0.09},
                  'is_frozen_trainable': False},
                'pipeline_1': {
                  'class': 'lale.operators.TrainablePipeline',
                  'state': 'trainable',
                  'edges': [['no_op', 'lr_1']],
                  'steps': {
                    'no_op': {
                      'class': 'lale.lib.lale.no_op.NoOpImpl',
                      'state': 'trained',
                      'operator': 'NoOp', 'label': 'NoOp',
                      'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.no_op.html',
                      'hyperparams': None,
                      'coefs': None,
                      'is_frozen_trainable': True, 'is_frozen_trained': True},
                    'lr_1': {
                      'class': 'lale.lib.sklearn.logistic_regression.LogisticRegressionImpl',
                      'state': 'trainable',
                      'operator': 'LogisticRegression', 'label': 'LR',
                      'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.logistic_regression.html',
                      'hyperparams': {'C': 0.19},
                      'is_frozen_trainable': False}}}}}}}
        json = to_json(operator)
        self.assertEqual(json, json_expected)
        operator_2 = from_json(json)
        json_2 = to_json(operator_2)
        self.assertEqual(json, json_2)

if __name__ == '__main__':
    unittest.main()



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sklearn.ensemble import VotingClassifier as SKLModel
import lale.docstrings
import lale.operators

class VotingClassifierImpl():
    def __init__(self, estimators=None, voting='hard', weights=None, n_jobs=None, flatten_transform=True):
        self._hyperparams = {
            'estimators': estimators,
            'voting': voting,
            'weights': weights,
            'n_jobs': n_jobs,
            'flatten_transform': flatten_transform}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self
    
    def transform(self, X):
        return self._wrapped_model.transform(X)

    def predict(self, X):
        return self._wrapped_model.predict(X)
    
    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Soft Voting/Majority Rule classifier for unfitted estimators.',
    'allOf': [{
        'type': 'object',
        'required': ['estimators', 'voting', 'weights', 'n_jobs', 'flatten_transform'],
        'relevantToOptimizer': ['voting'],
        'additionalProperties': False,
        'properties': {
            'estimators': {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'laleType': 'tuple',
                    'items': [
                        {'type':'string'},
                        {'laleType': 'operator'}
                    ]
                },
                'description': 'list of (string, estimator) tuples. Invoking the ``fit`` method on the ``VotingClassifier`` will fit clones'},
            'voting': {
                'enum': ['hard', 'soft'],
                'default': 'hard',
                'description': "If 'hard', uses predicted class labels for majority rule voting."},
            'weights': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }, {
                    'enum': [None]}],
                'default': None,
                'description': 'Sequence of weights (`float` or `int`) to weight the occurrences of'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of jobs to run in parallel for ``fit``.'},
            'flatten_transform': {
                'type': 'boolean',
                'default': True,
                'description': "Affects shape of transform output only when voting='soft'"},
        }}, {
        'description': "Parameter: flatten_transform > only when voting='soft' if voting='soft' and flatten_transform=true",
        'anyOf': [{
            'type': 'object',
            'properties': {
                'voting': {
                    'enum': ['soft']},
            }}, {
            'type': 'object',
            'properties': {
                'flatten_transform': {
                    'enum': [True]},
            }}]}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the estimators.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vectors, where n_samples is the number of samples and'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values.'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'description': 'Sample weights. If None, then samples are equally weighted.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return class labels or probabilities for X for each estimator.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vectors, where n_samples is the number of samples and'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': "If `voting='soft'` and `flatten_transform=True`:",
    'type': 'array',
    'items': {
        'type': 'array',
        'items': { 'anyOf':[{
            'type': 'number'},{
            'type': 'array',
            'items': {'type':'number'}
            }]}}}

_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted class labels.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Compute probabilities of possible outcomes for samples in X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples.'},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Weighted average probability for each class per sample.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}

_input_decision_function_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_output_decision_function_schema = {
    'description': 'Confidence scores for samples for each class in the model.',
    'anyOf': [
    {   'description': 'In the multi-way case, score per (sample, class) combination.',
        'type': 'array',
        'items': {'type': 'array', 'items': {'type': 'number'}}},
    {   'description': 'In the binary case, score for `self._classes[1]`.',
        'type': 'array',
        'items': {'type': 'number'}}]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Voting classifier`_ from scikit-learn for voting ensemble.

.. _`Voting classifier`: https://scikit-learn.org/0.20/modules/generated/sklearn.ensemble.VotingClassifier.html#sklearn-ensemble-votingclassifier
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.voting_classifier.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer', 'estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema,
        'input_decision_function': _input_decision_function_schema,
        'output_decision_function': _output_decision_function_schema,
}}

lale.docstrings.set_docstrings(VotingClassifierImpl, _combined_schemas)

VotingClassifier = lale.operators.make_operator(VotingClassifierImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.feature_selection
import lale.docstrings
import lale.operators
import pandas as pd

class SelectKBestImpl():
    def __init__(self, score_func=None, k=10):
        if score_func:
            self._hyperparams = {
                'score_func': score_func,
                'k': k}
        else:
            self._hyperparams = {
                'k': k
            }
        self._wrapped_model = sklearn.feature_selection.SelectKBest(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        if isinstance(X, pd.DataFrame):
            keep_indices = self._wrapped_model.get_support(indices=True)
            keep_columns = [X.columns[i] for i in keep_indices]
            result = X[keep_columns]
        else:
            result = self._wrapped_model.transform(X)
        return result

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Select features according to the k highest scores.',
    'allOf': [{
        'type': 'object',
        'required': ['score_func', 'k'],
        'relevantToOptimizer': ['k'],
        'additionalProperties': False,
        'properties': {
            'score_func':  {
                'anyOf': [{}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Function taking two arrays X and y, and returning a pair of arrays (scores, pvalues) or a single array with scores.'},
            'k': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 15}, {
                    'enum': ['all']}],
                'default': 10,
                'description': 'Number of top features to select'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Run score function on (X, y) and get the appropriate features.',
    'type': 'object',
    'required': ['X', 'y'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Training input samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {'type': 'number'}}},
        'y': {
            'type': 'array',
            'items': {'type': 'number'},
            'description': 'Target values.'}}}

_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Reduce X to the selected features.',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The input samples with only the selected features.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {'type': 'number'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Select k best`_ feature selection transformer from scikit-learn.

.. _`Select k best`: https://scikit-learn.org/0.20/modules/generated/sklearn.feature_selection.SelectKBest.html#sklearn-feature-selection-selectkbest
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.select_k_best.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(SelectKBestImpl, _combined_schemas)

SelectKBest = lale.operators.make_operator(SelectKBestImpl, _combined_schemas)




from sklearn.preprocessing.data import PolynomialFeatures as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class PolynomialFeaturesImpl():

    def __init__(self, degree=2, interaction_only=False, include_bias=True):
        self._hyperparams = {
            'degree': degree,
            'interaction_only': interaction_only,
            'include_bias': include_bias}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for PolynomialFeatures    Generate polynomial and interaction features.',
    'allOf': [{
        'type': 'object',
        'required': ['degree', 'interaction_only', 'include_bias'],
        'relevantToOptimizer': ['degree', 'interaction_only'],
        'additionalProperties': False,
        'properties': {
            'degree': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 3,
                'distribution': 'uniform',
                'default': 2,
                'description': 'The degree of the polynomial features'},
            'interaction_only': {
                'type': 'boolean',
                'default': False,
                'description': 'If true, only interaction features are produced: features that are products of at most ``degree`` *distinct* input features (so not ``x[1] ** 2``, ``x[0] * x[2] ** 3``, etc.).'},
            'include_bias': {
                'type': 'boolean',
                'default': True,
                'description': 'If True (default), then include a bias column, the feature in which all polynomial powers are zero (i.e'},
        }}, {
        'XXX TODO XXX': 'Parameter: interaction_only > only interaction features are produced: features that are products of at most degree *distinct* input features (so not x[1] ** 2, x[0] * x[2] ** 3, etc'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Compute number of output features.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The data.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform data to polynomial features',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix, shape [n_samples, n_features]'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'The data to transform, row by row'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The matrix of features, where NP is the number of polynomial features generated from the combination of inputs.',
    'laleType': 'Any',
    'XXX TODO XXX': 'np.ndarray or CSC sparse matrix, shape [n_samples, NP]',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.PolynomialFeatures#sklearn-preprocessing-polynomialfeatures',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(PolynomialFeaturesImpl, _combined_schemas)
PolynomialFeatures = lale.operators.make_operator(PolynomialFeaturesImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aif360.algorithms.postprocessing
import aif360.datasets
import aif360.metrics
import lale.datasets.data_schemas
import lale.type_checking
import pandas as pd

def dataset_to_pandas(dataset, return_only='Xy'):
    """
    Return pandas representation of the AIF360 dataset.

    Parameters
    ----------
    dataset : aif360.datasets.BinaryLabelDataset

      AIF360 dataset to convert to a pandas representation.

    return_only : 'Xy', 'X', or 'y'

      Which part of features X or labels y to convert and return.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe or None, features X

      - item 1: pandas Series or None, labels y
    """
    if 'X' in return_only:
        X = pd.DataFrame(dataset.features, columns=dataset.feature_names)
        result_X = lale.datasets.data_schemas.add_schema(X)
        assert isinstance(result_X, pd.DataFrame), type(result_X)
    else:
        result_X = None
    if 'y' in return_only:
        y = pd.Series(dataset.labels.ravel(), name=dataset.label_names[0])
        result_y = lale.datasets.data_schemas.add_schema(y)
        assert isinstance(result_y, pd.Series), type(result_y)
    else:
        result_y = None
    return result_X, result_y

_dataset_fairness_properties: lale.type_checking.JSON_TYPE = {
    'favorable_label': {
        'description': 'Label value which is considered favorable (i.e. "positive").',
        'type': 'number'},
    'unfavorable_label': {
        'description': 'Label value which is considered unfavorable (i.e. "negative").',
        'type': 'number'},
    'protected_attribute_names': {
        'description': 'Subset of feature names for which fairness is desired.',
        'type': 'array',
        'items': {'type': 'string'}},
    'unprivileged_groups': {
        'description': 'Representation for unprivileged group.',
        'type': 'array',
        'items': {
            'description': 'Map from feature names to group-indicating values.',
            'type': 'object',
            'additionalProperties': {
                'type': 'number'}}},
    'privileged_groups': {
        'description': 'Representation for privileged group.',
        'type': 'array',
        'items': {
            'description': 'Map from feature names to group-indicating values.',
            'type': 'object',
            'additionalProperties': {
                'type': 'number'}}}}

_dataset_fairness_schema = {
    'type': 'object',
    'properties': _dataset_fairness_properties}

def dataset_fairness_info(dataset):
    """
    Inspect the AIF360 dataset and return its fairness metadata as JSON.

    Parameters
    ----------
    dataset : aif360.datasets.BinaryLabelDataset

    Returns
    -------
    result : dict

      JSON data structure with fairness information.

      - favorable_label : number

          Label value which is considered favorable (i.e. "positive").

      - unfavorable_label : number

          Label value which is considered unfavorable (i.e. "negative").

      - protected_attribute_names : array **of** items : string

          Subset of feature names for which fairness is desired.

      - unprivileged_groups : array

          Representation for unprivileged group.

          - items : dict

              Map from feature names to group-indicating values.

      - privileged_groups : array

          Representation for privileged group.

          - items : dict

              Map from feature names to group-indicating values.
    """
    def attributes_to_groups(names, value_arrays):
        result = [{}]
        for i in range(len(names)):
            next_result = []
            for d in result:
                for next_v in value_arrays[i]:
                    next_d = {**d, names[i]: next_v}
                    next_result.append(next_d)
            result = next_result
        return result
    unprivileged_groups = attributes_to_groups(
        dataset.protected_attribute_names,
        dataset.unprivileged_protected_attributes)
    privileged_groups = attributes_to_groups(
        dataset.protected_attribute_names,
        dataset.privileged_protected_attributes)
    result = {
        'favorable_label': dataset.favorable_label,
        'unfavorable_label': dataset.unfavorable_label,
        'protected_attribute_names': dataset.protected_attribute_names,
        'unprivileged_groups': unprivileged_groups,
        'privileged_groups': privileged_groups}
    lale.type_checking.validate_schema(result, _dataset_fairness_schema)
    return result

class _PandasToDatasetConverter:
    def __init__(self, favorable_label, unfavorable_label, protected_attribute_names):
        lale.type_checking.validate_schema(favorable_label,
            _dataset_fairness_properties['favorable_label'])
        self.favorable_label = favorable_label
        lale.type_checking.validate_schema(unfavorable_label,
            _dataset_fairness_properties['unfavorable_label'])
        self.unfavorable_label = unfavorable_label
        lale.type_checking.validate_schema(protected_attribute_names,
            _dataset_fairness_properties['protected_attribute_names'])
        self.protected_attribute_names = protected_attribute_names

    def __call__(self, X, y):
        assert isinstance(X, pd.DataFrame), type(X)
        assert isinstance(y, pd.Series), type(y)
        assert X.shape[0] == y.shape[0], f'X.shape {X.shape}, y.shape {y.shape}'
        df = pd.concat([X, y], axis=1)
        assert not df.isna().any().any(), f'df\n{df}\nX\n{X}\ny\n{y}'
        label_names = [y.name]
        result = aif360.datasets.BinaryLabelDataset(
            favorable_label=self.favorable_label,
            unfavorable_label=self.unfavorable_label,
            protected_attribute_names=self.protected_attribute_names,
            df=df,
            label_names=label_names)
        return result

def _ensure_series(data, index, dtype, name):
    if isinstance(data, pd.Series):
        return data
    result = pd.Series(data=data, index=index, dtype=dtype, name=name)
    return result

class _BinaryLabelScorer:
    def __init__(self, metric, favorable_label, unfavorable_label, protected_attribute_names, unprivileged_groups, privileged_groups):
        assert hasattr(aif360.metrics.BinaryLabelDatasetMetric, metric)
        self.metric = metric
        self.fairness_info = {
            'favorable_label': favorable_label,
            'unfavorable_label': unfavorable_label,
            'protected_attribute_names': protected_attribute_names,
            'unprivileged_groups': unprivileged_groups,
            'privileged_groups': privileged_groups}
        lale.type_checking.validate_schema(
            self.fairness_info, _dataset_fairness_schema)
        self.pandas_to_dataset = _PandasToDatasetConverter(
            favorable_label, unfavorable_label, protected_attribute_names)

    def __call__(self, estimator, X, y):
        predicted = estimator.predict(X)
        y_pred = _ensure_series(predicted, X.index, y.dtype, y.name)
        dataset_pred = self.pandas_to_dataset(X, y_pred)
        fairness_metrics = aif360.metrics.BinaryLabelDatasetMetric(
            dataset_pred,
            self.fairness_info['unprivileged_groups'],
            self.fairness_info['privileged_groups'])
        method = getattr(fairness_metrics, self.metric)
        result = method()
        return result

def disparate_impact(favorable_label, unfavorable_label, protected_attribute_names, unprivileged_groups, privileged_groups):
    """
    Make a scikit-learn compatible scorer given the fairness info.

    Parameters
    ----------
    favorable_label : number

      Label value which is considered favorable (i.e. "positive").

    unfavorable_label : number

      Label value which is considered unfavorable (i.e. "negative").

    protected_attribute_names : array **of** items : string

      Subset of feature names for which fairness is desired.

    unprivileged_groups : array

      Representation for unprivileged group.

      - items : dict

          Map from feature names to group-indicating values.

    privileged_groups : array

      Representation for privileged group.

      - items : dict

          Map from feature names to group-indicating values.

    Returns
    -------
    result : callable

      Scorer that takes three arguments (estimator, X, y) and returns score.
    """
    return _BinaryLabelScorer('disparate_impact', favorable_label, unfavorable_label, protected_attribute_names, unprivileged_groups, privileged_groups)

def statistical_parity_difference(favorable_label, unfavorable_label, protected_attribute_names, unprivileged_groups, privileged_groups):
    """
    Make a scikit-learn compatible scorer given the fairness info.

    Parameters
    ----------
    favorable_label : number

      Label value which is considered favorable (i.e. "positive").

    unfavorable_label : number

      Label value which is considered unfavorable (i.e. "negative").

    protected_attribute_names : array **of** items : string

      Subset of feature names for which fairness is desired.

    unprivileged_groups : array

      Representation for unprivileged group.

      - items : dict

          Map from feature names to group-indicating values.

    privileged_groups : array

      Representation for privileged group.

      - items : dict

          Map from feature names to group-indicating values.

    Returns
    -------
    result : callable

      Scorer that takes three arguments (estimator, X, y) and returns score.
    """
    return _BinaryLabelScorer('statistical_parity_difference', favorable_label, unfavorable_label, protected_attribute_names, unprivileged_groups, privileged_groups)

_postprocessing_base_hyperparams = {
    'estimator': {
        'description': 'Nested supervised learning operator for which to mitigate fairness.',
        'laleType': 'estimator'},
    'favorable_label': _dataset_fairness_properties['favorable_label'],
    'unfavorable_label': _dataset_fairness_properties['unfavorable_label'],
    'protected_attribute_names': _dataset_fairness_properties['protected_attribute_names']}

class _BasePostprocessingImpl:
    def __init__(self, mitigator, estimator, favorable_label, unfavorable_label, protected_attribute_names):
        self.mitigator = mitigator
        self.estimator = estimator
        self.pandas_to_dataset = _PandasToDatasetConverter(
            favorable_label, unfavorable_label, protected_attribute_names)
        self.y_dtype = None
        self.y_name = None

    def fit(self, X, y):
        self.y_dtype = y.dtype
        self.y_name = y.name
        y_true = y
        self.estimator = self.estimator.fit(X, y_true)
        predicted = self.estimator.predict(X)
        y_pred = _ensure_series(predicted, X.index, self.y_dtype, self.y_name)
        dataset_true = self.pandas_to_dataset(X, y_true)
        dataset_pred = self.pandas_to_dataset(X, y_pred)
        self.mitigator = self.mitigator.fit(dataset_true, dataset_pred)
        return self

    def predict(self, X):
        predicted = self.estimator.predict(X)
        y_pred = _ensure_series(predicted, X.index, self.y_dtype, self.y_name)
        dataset_pred = self.pandas_to_dataset(X, y_pred)
        dataset_out = self.mitigator.predict(dataset_pred)
        _, y_out = dataset_to_pandas(dataset_out, return_only='y')
        return y_out

_numeric_supervised_input_fit_schema = {
    'type': 'object',
    'required': ['X', 'y'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {'type': 'number'}}},
        'y': {
            'description': 'Target class labels; the array is over samples.',
            'type': 'array',
            'items': {'type': 'number'}}}}

_numeric_input_predict_schema = {
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_numeric_output_predict_schema = {
    'description': 'Predicted class label per sample.',
    'type': 'array', 'items': {'type': 'number'}}



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

import lale.lib.sklearn
import lale.search.lale_grid_search_cv
import lale.operators
import lale.sklearn_compat
from .observing import Observing
from .observing import ObservingImpl
# from abc import ABC

from typing import Any, Dict

class GridSearchCVImpl:
    def __init__(self, estimator=None, cv=5, scoring='accuracy', n_jobs=None, lale_num_samples=None, lale_num_grids=None, param_grid=None, pgo=None, observer=None):
        if observer is not None and isinstance(observer, type):
            # if we are given a class name, instantiate it
            observer = observer()

        self._hyperparams = {
            'estimator': estimator,
            'cv': cv,
            'scoring': scoring,
            'n_jobs': n_jobs,
            'lale_num_samples': lale_num_samples,
            'lale_num_grids': lale_num_grids,
            'pgo': pgo,
            'hp_grid': param_grid,
            'observer': observer }

    def fit(self, X, y):
        if self._hyperparams['estimator'] is None:
            op = lale.lib.sklearn.LogisticRegression
        else:
            op = self._hyperparams['estimator']

        observed_op = op
        obs = self._hyperparams['observer']
        if obs is not None:
            observed_op = Observing(op=op, observer=obs)

        hp_grid = self._hyperparams['hp_grid']
        if hp_grid is None:
            hp_grid = lale.search.lale_grid_search_cv.get_parameter_grids(
                observed_op,
                num_samples=self._hyperparams['lale_num_samples'],
                num_grids=self._hyperparams['lale_num_grids'],
                pgo=self._hyperparams['pgo'])
        if not hp_grid and isinstance(op, lale.operators.IndividualOp):
            hp_grid = [
                lale.search.lale_grid_search_cv.get_defaults_as_param_grid(observed_op)]
        if hp_grid:
            if obs is not None:
                observed_op._impl.startObserving("optimize", hp_grid=hp_grid, op=op,
                    num_samples=self._hyperparams['lale_num_samples'],
                    num_grids=self._hyperparams['lale_num_grids'],
                    pgo=self._hyperparams['pgo'])
            try:
                self.grid = lale.search.lale_grid_search_cv.get_lale_gridsearchcv_op(
                    lale.sklearn_compat.make_sklearn_compat(observed_op),
                    hp_grid,
                    cv=self._hyperparams['cv'],
                    scoring=self._hyperparams['scoring'],
                    n_jobs=self._hyperparams['n_jobs'])
                self.grid.fit(X, y)
                be = self.grid.best_estimator_.to_lale()
            except BaseException as e:
                if obs is not None:
                    assert isinstance(be._impl, ObservingImpl)
                    observed_op._impl.failObserving("optimize", e)
                raise

            if obs is not None:
                assert isinstance(be._impl, ObservingImpl)
                be_base = be._impl.getOp()
                observed_op._impl.endObserving("optimize", best=be)
            self._best_estimator = be
        else:
            assert isinstance(op, lale.operators.TrainableOperator)
            self._best_estimator = op
        return self

    def predict(self, X):
        return self._best_estimator.predict(X)

    def get_pipeline(self, pipeline_name=None, astype='lale'):
        if pipeline_name is not None:
            raise NotImplementedError('Cannot get pipeline by name yet.')
        result = getattr(self, '_best_estimator', None)
        if result is None or astype == 'lale':
            return result
        assert astype == 'sklearn', astype
        return lale.sklearn_compat.make_sklearn_compat(result)

_hyperparams_schema = {
    'allOf': [
    {   'type': 'object',
        'required': [
            'estimator', 'cv', 'scoring', 'n_jobs', 'lale_num_samples',
            'lale_num_grids', 'pgo'],
        'relevantToOptimizer': ['estimator'],
        'additionalProperties': False,
        'properties': {
            'estimator': {
                'description': 'Planned Lale individual operator or pipeline,\nby default LogisticRegression.',
                'anyOf': [
                {   'laleType': 'operator',
                    'not': {'enum': [None]}},
                {   'enum': [None],
                    'description': 'lale.lib.sklearn.LogisticRegression'}],
                'default': None},
            'cv': {
                'description': 'Number of folds for cross-validation.',
                'type': 'integer',
                'minimum': 1,
                'default': 5},
            'scoring': {
                'description': 'Scorer object, or known scorer named by string.',
                'anyOf': [
                {   'description': 'Custom scorer object, see https://scikit-learn.org/stable/modules/model_evaluation.html',
                    'not': {'type': 'string'}},
                {   'description': 'Known scorer for classification task.',
                    'enum': [
                        'accuracy', 'explained_variance', 'max_error',
                        'roc_auc', 'roc_auc_ovr', 'roc_auc_ovo',
                        'roc_auc_ovr_weighted', 'roc_auc_ovo_weighted',
                        'balanced_accuracy', 'average_precision',
                        'neg_log_loss', 'neg_brier_score']},
                {   'description': 'Known scorer for regression task.',
                    'enum': [
                        'r2', 'neg_mean_squared_error',
                        'neg_mean_absolute_error',
                        'neg_root_mean_squared_error',
                        'neg_mean_squared_log_error',
                        'neg_median_absolute_error']}],
                'default': 'accuracy'},
            'n_jobs': {
                'description': 'Number of jobs to run in parallel.',
                'anyOf': [
                    {   'description':
                            '1 unless in joblib.parallel_backend context.',
                        'enum': [None]},
                    {   'description': 'Use all processors.',
                        'enum': [-1]},
                    {   'description': 'Number of jobs to run in parallel.',
                        'type': 'integer',
                        'minimum': 1}],
                'default': None},
            'lale_num_samples': {
                'description': 'How many samples to draw when discretizing a continuous hyperparameter.',
                'anyOf': [
                    {   'type': 'integer',
                        'minimum': 1},
                    {   'description': 'lale.search.lale_grid_search_cv.DEFAULT_SAMPLES_PER_DISTRIBUTION',
                        'enum': [None]}],
                'default': None},
            'lale_num_grids': {
                'description': 'How many top-level disjuncts to explore.',
                'anyOf': [
                {   'description': 'If not set, keep all grids.',
                    'enum': [None]},
                {   'description': 'Fraction of grids to keep.',
                    'type': 'number',
                    'minimum': 0.0,
                    'exclusiveMinimum': True,
                    'maximum': 1.0,
                    'exclusiveMaximum': True},
                {   'description': 'Number of grids to keep.',
                    'type': 'integer',
                    'minimum': 1}],
                'default': None},
            'param_grid': {
                'anyOf':[
                {
                'enum': [None],
                'description': 'Generated automatically.'},
                {'description': 'Dictionary of hyperparameter ranges in the grid.'}],
                'default': None},
            'pgo': {
                'anyOf': [
                {   'description': 'lale.search.PGO'},
                {   'enum': [None]}],
                'default': None},
            'observer': {
                'laleType':'Any',
                'default': None,
                'description': 'a class or object with callbacks for observing the state of the optimization'}
                }}]}

_input_fit_schema = {
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {},
        'y': {}}}

_input_predict_schema = {
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {}}}

_output_predict_schema:Dict[str,Any] = {}

_combined_schemas = {
    'description': """GridSearchCV_ performs an exhaustive search over a discretized space.

.. _GridSearchCV: https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.hyperopt_classifier.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema}}

lale.docstrings.set_docstrings(GridSearchCVImpl, _combined_schemas)

GridSearchCV = lale.operators.make_operator(GridSearchCVImpl, _combined_schemas)
